#
#    Copyright (c) 2009 Tom Keffer <tkeffer@gmail.com>
#
#    See the file LICENSE.txt for your full rights.
#
"""Various handly utility routines that don't belong anywhere else.

"""

import datetime
import time
import math
import ImageFont
import configobj

def get_font_handle(fontpath, *args):
    
    font = None
    if fontpath is not None :
        try :
            if fontpath.endswith('.ttf'):
                font = ImageFont.truetype(fontpath, *args)
            else :
                font = ImageFont.load_path(fontpath)
        except IOError :
            pass
    
    if font is None :
        font = ImageFont.load_default()
        
    return font 


def accumulatescalars(d):
    """Merges scalar options above a ConfigObj section with itself, accumulating the results.
    
    This routine is useful for specifying defaults near the root node, 
    then having them overridden in the leaf nodes of a ConfigObj.
    
    d: instance of a configobj.Section (i.e., a section of a ConfigObj)
    """
    
    # Use recursion. If I am the root object, then there is nothing above 
    # me to accumulate. Start with a virgin ConfigObj
    if d.parent is d :
        cum_dict = configobj.ConfigObj()
    else :
        # Otherwise, recursively accumulate scalars above me
        cum_dict = accumulatescalars(d.parent)
        
    # Now merge my scalars into the results:
    merge_dict = {}
    for k in d.scalars :
        merge_dict[k] = d[k]
    cum_dict.merge(merge_dict)
    return cum_dict

def stampgen(startstamp, stopstamp, interval):
    """Generator function yielding a sequence of timestamps, spaced interval apart.
    
    The sequence will fall on the same local time boundary as startstamp. 
    For example, if startstamp is epoch time 1236560400 (local time 
    8-Mar-2009 18:00 PST) and interval is 10800 (3 hours), then the yielded sequence 
    will be (shown with local times):
    
    1236560400 (8-Mar-2009 18:00 PST)
    1236571200 (8-Mar-2009 21:00 PST)
    1236582000 (9-Mar-2009 00:00 PST)
    1236592800 (9-Mar-2009 03:00 PDT)
    1236603600 (9-Mar-2009 06:00 PDT), etc.
    
    Note that DST started in the middle of the sequence and that therefore the
    actual time deltas between stamps is not necessarily 3 hours.
    
    startstamp: The start of the sequence in unix epoch time.
    
    stopstamp: The end of the sequence in unix epoch time. 
    
    interval: The time length of an interval in seconds.
    
    yields a sequence of timestamps between startstamp and endstamp, inclusive.
    """
    dt      = datetime.datetime.fromtimestamp(startstamp)
    stop_dt = datetime.datetime.fromtimestamp(stopstamp)
    if interval == 365.25/12 * 24 * 3600 :
        # Interval is a nominal month. This algorithm is 
        # necessary because not all months have the same length.
        while dt <= stop_dt :
            t_tuple = dt.timetuple()
            yield time.mktime(t_tuple)
            year  = t_tuple[0]
            month = t_tuple[1]
            month += 1
            if month > 12 :
                month -= 12
                year  += 1
            dt = dt.replace(year=year, month=month)
    else :
        # This rather complicated algorithm is necessary (rather than just
        # doing some time stamp arithmetic) because of the possibility that DST
        # changes in the middle of an interval
        delta  = datetime.timedelta(seconds=interval)
        while dt <= stop_dt :
            yield int(time.mktime(dt.timetuple()))
            dt += delta

def intervalgen(start_ts, stop_ts, interval):
    """Generator function yielding a sequence of time intervals.
    
    Yields a sequence of intervals. First interval is (start_ts, start_ts+interval),
    second is (start_ts+interval, start_ts+2*interval), etc. The last interval
    will end at or before stop_ts. It is up to the consumer to interpret whether
    the end points of any given interval is inclusive or exclusive to the
    interval.
    
    start_ts: The start of the first interval in unix epoch time.
    
    stop_ts: The end of the last interval will be equal to or less than this.
    In unix epoch time.
    
    interval: The time length of an interval in seconds.
    
    yields: A sequence of 2-tuples. First value is start of the interval, second 
    is the end. Both the start and end will be on the same time boundary as
    start_ts
    
    """  
    dt1     = datetime.datetime.fromtimestamp(start_ts)
    stop_dt = datetime.datetime.fromtimestamp(stop_ts)
    
    if interval == 365.25/12 * 24 * 3600 :
        # Interval is a nominal month. This algorithm is 
        # necessary because not all months have the same length.
        while dt1 < stop_dt :
            t_tuple = dt1.timetuple()
            year  = t_tuple[0]
            month = t_tuple[1]
            month += 1
            if month > 12 :
                month -= 12
                year  += 1
            dt2 = min(dt1.replace(year=year, month=month), stop_dt)
            stamp1 = time.mktime(t_tuple)
            stamp2 = time.mktime(dt2.timetuple())
            yield (stamp1, stamp2)
            dt1 = dt2
    else :
        # This rather complicated algorithm is necessary (rather than just
        # doing some time stamp arithmetic) because of the possibility that DST
        # changes in the middle of an interval
        delta  = datetime.timedelta(seconds=interval)
        while dt1 < stop_dt :
            dt2 = min(dt1 + delta, stop_dt)
            stamp1 = time.mktime(dt1.timetuple())
            stamp2 = time.mktime(dt2.timetuple())
            yield (stamp1, stamp2)
            dt1 = dt2

def monthRange(start_ts, stop_ts):
    """Generator function that generates start/stop of months in an
    inclusive range.
    
    For example, if start_ts is 1196705700 (local time 2007-12-03 10:15:00 PST)
    and stop_ts is 1206101100 (2008-03-21, 05:05:00 PST), the function
    will generate:
     
    1196496000 (2007-12-01 00:00:00 PST) 1199174400 (2008-01-01 00:00:00 PST) 
    1199174400 (2008-01-01 00:00:00 PST) 1201852800 (2008-02-01 00:00:00 PST)
    1201852800 (2008-02-01 00:00:00 PST) 1204358400 (2009-03-01 00:00:00 PST)
    1204358400 (2009-03-01 00:00:00 PST) 1207033200 (2009-04-01 00:00:00 PDT)
    
    Note that a daylight savings time change happened 8 March 2009.

    start_ts: A time stamp somewhere in the first month.
    
    stop_ts: A time stamp somewhere in the last month.
    
    yields: Tuples (start_month, start_next_month). start_month
    is the time stamp of the start of the month. start_next_month
    is the time stamp of the start of the next month.
    
    """
    _start_dt = datetime.date.fromtimestamp(start_ts)
    _stop_dt  = datetime.date.fromtimestamp(stop_ts)
    
    _this_month_d = _start_dt.replace(day=1)
    _stop_month_d = _stop_dt.replace(day=1)
    
    _this_month_ts = time.mktime(_this_month_d.timetuple())

    while _this_month_d <= _stop_month_d :
        _yr = _this_month_d.year
        _mo = _this_month_d.month
        if _mo == 12 :
            _yr += 1
            _mo = 1
        else :
            _mo += 1
        _next_month_d = datetime.date(_yr, _mo, 1)
        _next_month_ts = time.mktime(_next_month_d.timetuple())
        yield(_this_month_ts, _next_month_ts)
        _this_month_d = _next_month_d
        _this_month_ts = _next_month_ts

def startOfDay(time_ts):
    """Calculate the unix epoch time for the start of a (local time) day.
    
    time_ts: A timestamp somewhere in the day for which the start-of-day
    is desired.
    
    returns: The timestamp for the start-of-day (00:00) in unix epoch time.
    
    """
    _time_tt = time.localtime(time_ts)
    _bod_ts  = time.mktime((_time_tt.tm_year, 
                            _time_tt.tm_mon,
                            _time_tt.tm_mday,
                            0, 0, 0, 0, 0, -1))
    return int(_bod_ts)

def startOfArchiveDay(time_ts):
    """Given an archive time stamp, calculate its start of day.
    
    similar to startOfDay(), except that an archive stamped at midnight
    actually belongs to the *previous* day.

    time_ts: A timestamp somewhere in the day for which the start-of-day
    is desired.
    
    returns: The timestamp for the start-of-day (00:00) in unix epoch time.
    
    """
    _time_dt = datetime.datetime.fromtimestamp(time_ts)
    if (_time_dt.hour, _time_dt.minute, _time_dt.second) == (0, 0, 0):
        _time_dt -= datetime.timedelta(days=1)
    _time_tt = _time_dt.timetuple()
    _bod_ts  = time.mktime((_time_tt.tm_year, 
                            _time_tt.tm_mon,
                            _time_tt.tm_mday,
                            0, 0, 0, 0, 0, -1))
    return int(_bod_ts)

    
def startOfWeek(time_ts):
    """Calculate the unix epoch time of the start of a week.
    
    time_ts: A timestamp somewhere in the week for which the start-of-week
    is desired.
    
    returns: The timestamp for the start-of-the-week (00:00 Sunday) in
    unix epoch time.
    """
    _time_dt = datetime.datetime.fromtimestamp(time_ts)
    _day_of_week = _time_dt.weekday()
    _delta = _day_of_week + 1 if _day_of_week is not 6 else 0
    _delta_td = datetime.timedelta(days = _delta)
    _sunday_dt = _time_dt - _delta_td
    _sunday_tt = _sunday_dt.timetuple()
    _bow_ts = time.mktime((_sunday_tt.tm_year,
                           _sunday_tt.tm_mon,
                           _sunday_tt.tm_mday,
                            0, 0, 0, 0, 0, -1))
    return int(_bow_ts)
    
def startOfMonth(time_ts):
    """Calculate the unix epoch time for the start of a (local time) month.
    
    time_ts: A timestamp somewhere in the month for which the start-of-month
    is desired.
    
    returns: The timestamp for the start-of-month (YYYY-MM-01 00:00) in
    unix epoch time.
    
    """
    _time_tt = time.localtime(time_ts)
    _bom_ts  = time.mktime((_time_tt.tm_year, 
                            _time_tt.tm_mon,
                            1,
                            0, 0, 0, 0, 0, -1))
    return int(_bom_ts)

def startOfYear(time_ts):
    """Calculate the unix epoch time for the start of a (local time) year.
    
    time_ts: A timestamp somewhere in the year for which the start-of-year
    is desired.
    
    returns: The timestamp for the start-of-year (YYYY-01-01 00:00) in
    unix epoch time.
    
    """
    _time_tt = time.localtime(time_ts)
    _boy_ts  = time.mktime((_time_tt.tm_year, 
                            1,
                            1,
                            0, 0, 0, 0, 0, -1))
    return int(_boy_ts)

def startOfRainYear(time_ts, sory_mon=1):
    """Calculate the unix epoch time for the start of a (local time) rain year.
    
    time_ts: A timestamp somewhere in the rain year for which the start-of-rain-year
    is desired.
    
    sory_mon: The month that the rain year starts.
    
    returns: The timestamp for the start-of-rain-year in unix epoch time.
    
    """
    _time_tt = time.localtime(time_ts)
    # If we have not passed the month that the rain year starts, then the
    # calendar year is last year, otherwise this year. 
    year = _time_tt.tm_year if _time_tt.tm_mon >= sory_mon else _time_tt.tm_year - 1
    
    _bory_ts  = time.mktime((year, 
                            sory_mon,
                            1,
                            0, 0, 0, 0, 0, -1))
    return int(_bory_ts)

def secs_to_string(secs):
    """Convert seconds to a string with days, hours, minutes and seconds"""
    str_list = []
    for (label, interval) in (('day', 86400), ('hour', 3600), ('minute', 60), ('second', 1)):
        amt = int(secs / interval)
        plural = '' if amt == 1 else 's'
        str_list.append("%d %s%s" % (amt, label, plural))
        secs %= interval
    str = ', '.join(str_list)
    return str

def latlon_string(ll, hemi):
    """Format a decimal latitude or longitude as a string"""
    labs = abs(ll)
    (frac, deg) = math.modf(labs)
    min = frac * 60.0
    str = "%d\xb0%0.2f'%s" % (deg, min, hemi)
    return str.decode('string_escape')

if __name__ == '__main__':
    start_ts = time.mktime((2007, 12, 3, 10, 15, 0, 0, 0, -1))
    stop_ts  = time.mktime((2008,  3, 21, 5, 5,  0, 0, 0, -1))
    print start_ts, stop_ts
    
    for month in monthRange(start_ts, stop_ts):
        print "start: %s (%d); stop: %s (%d)" % (time.ctime(month[0]), month[0], time.ctime(month[1]), month[1])
        