#
#    Copyright (c) 2009 Tom Keffer <tkeffer@gmail.com>
#
#    See the file LICENSE.txt for your full rights.
#
"""Simulate a VantagePro interface, using a database.

"""
import time
import syslog
import weewx
import weewx.archive

class SimStation(object):
    """The strategy depends on their being a backing database with more data
    in it than the target database. The SimStation reads from the backing
    database, simulating the arrival of new packets.
    """
    
    def __init__(self, config_dict, start_time):
        """Initialize an instance of SimStation.
        
        config_dict: A configuration dictionary. The filename of
        the backing database will be extracted from 
        key ['SimStation']['sim_db'].
        
        start_time: This time will be the current time on the simulator.
        If you were to ask for the latest packet in the simulator, you
        would get this time. 
        """
        self.time_ts = int(start_time)
        self._sim_db = weewx.archive.Archive(config_dict['SimStation']['sim_db'])
        self.interval = 60 * config_dict['SimStation'].as_int('archive_interval')

        
    def getTime(self):
        """Unfortunately, the conceit isn't complete. We simply return the present
        time.
        """
        return time.localtime()
        
    def setTime(self, time_tt):
        """Set the time on the simulator. Ignored.
        """
        syslog.syslog(syslog.LOG_NOTICE, "SimStation: Time set to %s (%d)" % (time.asctime(time_tt), time.mktime(time_tt)))
        
    
    def simNewPacket(self):
        """Pretend a new packet has come in.
        """
        self.time_ts += self.interval
        syslog.syslog(syslog.LOG_INFO, "SimStation: added new packet %s" % time.ctime(self.time_ts))

    
    def getArchivePackets(self, since_time):
        """A generator function that gets and return archive packets from a SimStation station.
        
        since_time: A time-tuple. All data since (but not including) this time will be returned.
        Pass in None for all data
        
        yields: a sequence of ArchivePackets containing the data
        """
        
        if since_time :
            since_time_ts = int(time.mktime(since_time))
        else :
            # Return 2 weeks worth of data (about what the VP holds in its
            # internal archive).
            since_time_ts = self.time_ts - 14 * 24 * 3600  
            
        return self._sim_db.getBatchRecords(since_time_ts, self.time_ts)

    def getArchivePacketsTS(self, timestamp) :
        """A generator function that gets and return archive data from a SimStation.
        
        timestamp: An epoch time. All data since (but not including) this time will be returned.
        Pass in None for all data
        
        yields: a sequence of ArchivePackets containing the data
        """
        
        time_tuple = time.localtime(timestamp) if timestamp else None
        
        return self.getArchivePackets(time_tuple)

        
    def getRX(self):
        return (12345, 134, 1, 22, 0)
