#
#    Copyright (c) 2009 Tom Keffer <tkeffer@gmail.com>
#
#    See the file LICENSE.txt for your full rights.
#

"""This module provides various default factories. It is tried if there is no user supplied
factory module."""

import weewx.stationdata
import weewx.genfiles
import weewx.genimages
import weewx.ftpdata
import weewx.processthread

def dataFactory(config_dict):
    """This factory function should return an object that will hold relevant 
    station data. Typically, it will be passed to the constructor of any generation 
    objects. The default version returns an instance of weewx.stationdata.StationData,
    which includes a fairly comprehensive set of data. It can easily be specialized 
    if you want to include any special data for special processing."""
    return weewx.stationdata.StationData(config_dict)

def genFilesFactory(config_dict, stationData):
    """This factory function should return an object that will process file-based
    products, such as HTML, or the NOAA reports. The default version returns an
    instance of weewx.genfiles.FileGen. You should return a specialized version instead
    if you want to do any special file processing. Its method 'run' will be called
    by the main processing thread."""
    return weewx.genfiles.GenFiles(config_dict, stationData)

def genImagesFactory(config_dict, stationData):
    """This factory function should return an object that will generate weather
    plots. The default version returns an instance of weewx.imagegen.ImageGen.
    You should return a specialized version instead if you want to generate any
    special images. Its method 'run' will be called by the main processing
    thread."""
    return weewx.genimages.GenImages(config_dict, stationData)

def ftpFactory(config_dict):
    """This factory function should return an object that will FTP the contents
    of the HTML directory to your web server. The default version returns an
    instance of weewx.ftp.FtpData. Either return an instance of your own class
    or specialize class FtpData to suit your needs. Its method 'run' will be
    called by the main processing thread."""
    return weewx.ftpdata.FtpData(config_dict)

def processThreadFactory(config_dict):
    """This factory function should return a specialized version of threading.Thread,
    and should do the main data processing in its method 'run'."""
    return weewx.processthread.ProcessThread(config_dict)