#
#    Copyright (c) 2009 Tom Keffer <tkeffer@gmail.com>
#
#    See the file LICENSE.txt for your full rights.
#

import os.path
import time
import syslog

import Cheetah.Template

import weewx.formatter
import weeutil.Almanac

class GenFiles(object):

    def __init__(self, config_dict, stationData):
        
        self.stationData = stationData
        
        # Get a formatter:
        self.formatter = self.getFormatter(config_dict)

        # Get a view into the stats.
        self.statsView = self.getModelView(config_dict)
    
        # Get an appropriate Almanac:
        self.almanac = self.getAlmanac(config_dict)

        # Get the directory holding the templates
        self.template_dir = os.path.join(config_dict['Station']['WEEWX_ROOT'],
                                         config_dict['HTML'].get('template_root', 'templates'))
        # Get the directories that will hold the generated HTML files:
        self.html_dir     = os.path.join(config_dict['Station']['WEEWX_ROOT'],
                                         config_dict['HTML'].get('html_root', 'public_html'))
    

    def getFormatter(self, config_dict):
        """Return a Formatter object. Can be overridden by subclasses."""
        return weewx.formatter.Formatter(config_dict['Labels']['ImperialFormats'],
                                         config_dict['Labels']['ImperialUnits'],
                                         config_dict['Labels']['Time'])

    def getModelView(self, config_dict):
        """Return a ModelView object. Can be overridden by subclasses."""
        return weewx.formatter.ModelView(self.stationData.stats, self.formatter)
    
    def getAlmanac(self, config_dict):
        """Return an Almanac. Can be overridden by subclasses."""
        return weeutil.Almanac.Almanac(self.stationData.last_timestamp, 
                                       self.stationData.station.latitude_f, self.stationData.station.longitude_f,
                                       config_dict['Almanac'].get('moon_phases', None))
    
    def genHtml(self, time_ts = None):
        """Generate HTML pages for a given timestamp.
        
        time_ts: The time for which the HTML pages are to be generated.
        [optional. Default is to generate pages for the last good timestamp
        in the database archive.]
        
        """
        ngen = 0
        t1 = time.time()
    
        if not time_ts:
            time_ts = self.stationData.last_timestamp
            
        for timescale in ('index', 'week', 'month', 'year'):
    
            template_file = os.path.join(self.template_dir, timescale + ".tmpl")
    
            html = Cheetah.Template.Template(file = template_file, 
                                             searchList = [{'station' : self.stationData.station, 
                                                            'almanac' : self.almanac},
                                                            self.statsView])
    
            file = open(os.path.join(self.html_dir, timescale + ".html"), mode='w')
            print >>file, html
            ngen += 1
    
        t2 = time.time()
        elapsed_time = t2-t1
        syslog.syslog(syslog.LOG_INFO, "genfiles: generated %d HTML pages in %.3f seconds" % (ngen, elapsed_time))

    def run(self):
        self.genHtml()
