#
#    Copyright (c) 2009 Tom Keffer <tkeffer@gmail.com>
#
#    See the file LICENSE.txt for your full rights.
#
"""Create NOAA monthly summary files.

"""

import time
import os.path
import calendar

import Cheetah.Template

import weeutil.weeutil
import weewx.stats
import weewx.html

def NoaaGen(config_dict, start_ts, stop_ts):
    """Generate monthly NOAA reports for the months between two time stamps."""

    # Get the directory holding the templates
    template_dir = os.path.join(config_dict['Station']['WEEWX_ROOT'],
                                config_dict['HTML'].get('template_root', 'templates'))

    # Get the template to be used for the NOAA reports:
    noaa_template = os.path.join(template_dir, 'NOAA.tmpl')
    
    # Get the directory that will hold the generated NOAA reports:
    noaa_root = os.path.join(config_dict['Station']['WEEWX_ROOT'],
                             config_dict['NOAA'].get('noaa_root', 'public_html/NOAA'))

    # Extract the station data from the configuration dictionary:
    station = weewx.html.Station(config_dict)
    
    # Open up the stats database
    statsFilename = os.path.join(config_dict['Station']['WEEWX_ROOT'], 
                                 config_dict['Stats']['stats_file'])
    statsDb = weewx.stats.StatsDb(statsFilename)
    
    # This is the start-of-day of the first timestamp:
    _sod_start_ts = weeutil.weeutil.startOfDay(start_ts)
    
    # Cycle through all the months inclusive in the start and stop times:
    for (_month_start_ts, _month_stop_ts) in weeutil.weeutil.monthRange(start_ts, stop_ts):
        print "Month start = %s(%d); stop = %s(%d)" %(time.ctime(_month_start_ts), _month_start_ts, 
                                                      time.ctime(_month_stop_ts),  _month_stop_ts)
        # Calculate the file name for this month
        _month_start_tt = time.localtime(_month_start_ts)
        (_yr, _mo) = _month_start_tt[0:2]
        _filename = "NOAA-%d-%d.txt" % (_yr, _mo)
        _fullpath = os.path.join(noaa_root, _filename)
        # If the file doesn't exist, or this is the final month, then we must generate the file:
        if not os.path.exists(_fullpath) or _month_start_ts < stop_ts <= _month_stop_ts:
            print "Need to generate %s." %(_fullpath,)

            # Print a suitable header for the month:
            monthName = time.strftime("%b", _month_start_tt)
            print "monthName=", monthName
            text = Cheetah.Template.Template(file = noaa_template, 
                                             searchList = [{'station' : station, 
                                                            'year'    : _yr, 
                                                            'month'   : monthName}])
            file = open(_fullpath, mode='w')
            print >>file, text
            
            # Now go through each day in the month
            (_wd, _days_in_month) = calendar.monthrange(_yr, _mo)
            for _day in xrange(_days_in_month):
                _sod = time.mktime((_yr, _mo, _day+1, 0, 0, 0, 0, 0, -1))
                if _sod < _sod_start_ts: continue
                if _sod > stop_ts: break
                print "%s, (%d)" % (time.ctime(_sod), _sod)
            
    
    
if __name__ == '__main__':
    import os.path
    import configobj
    import weewx.archive
    
    def main(config_path):
        try :
            config_dict = configobj.ConfigObj(config_path, file_error=True)
        except IOError:
            print "Unable to open configuration file ", config_path
            exit()
    
        archive = weewx.archive.Archive(os.path.join(config_dict['Station']['WEEWX_ROOT'], 
                                                     config_dict['Archive']['archive_file']))
        
        start_ts = archive.firstGoodStamp()
        stop_ts  = archive.lastGoodStamp()
        print "start = %s(%d); stop=%s(%d)" % (time.ctime(start_ts), start_ts, time.ctime(stop_ts), stop_ts)
        NoaaGen(config_dict, start_ts, stop_ts)
        
    main('../weewx.conf')