#
#    Copyright (c) 2009 Tom Keffer <tkeffer@gmail.com>
#
#    See the file LICENSE.txt for your full rights.
#
"""Defines the default station data, available for processing data.
"""
import os.path
import time

import weewx
import weewx.archive
import weewx.stats
import weewx.formatter
import weeutil.Almanac
import weeutil.weeutil

class StationData(object):
    '''
    Provides a lot of useful
    data for processing HTML, doing image generation, and various reports.
    '''
    def __init__(self, config_dict):
        """
        Initialize an instance of DefaultWorker. 
        """
        self.config_dict = config_dict
        
        # (Mostly) static data about the station. Seldom changes.
        self.station = Station(config_dict)
        
        # Open up the main database archive
        archiveFilename = os.path.join(config_dict['Station']['WEEWX_ROOT'], 
                                       config_dict['Archive']['archive_file'])
        self.archive = weewx.archive.Archive(archiveFilename)
        self.last_timestamp = self.archive.lastGoodStamp()
        
        # Open up the stats database
        statsFilename = os.path.join(config_dict['Station']['WEEWX_ROOT'], 
                                     config_dict['Stats']['stats_file'])
        self.statsDb = weewx.stats.StatsDb(statsFilename)
            
        # Statistics (hi, low, avg, etc.) will be held in this dictionary,
        # keyed by the time period ('current', 'day', 'month', 'year', and 'rainyear') 
        self.stats = {}
    
        # Get the 'current' conditions from it:
        self.stats['current'] = self.archive.getRecord(self.last_timestamp)
    
        # Get 'day', 'month', 'year', and 'rainyear' statistics from the stats database:
        self.stats['day'], self.stats['week'], self.stats['month'], self.stats['year'], self.stats['rainyear'] = \
                self.statsDb.getAllStats(self.last_timestamp, self.station.rain_year_start)


        
class Station(object):
    """Static data about the station. Rarely changes.
    """
    def __init__(self, config_dict):
        """Extracts info from the config_dict and stores it in self.
        
        """
        # TODO: Should allow I18N of 'N', 'S', etc. labels
        self.latitude_f  = config_dict['Station'].as_float('latitude')
        self.latitude    = weeutil.weeutil.latlon_string(self.latitude_f, 'N' if self.latitude_f>0 else 'S')
        self.longitude_f = config_dict['Station'].as_float('longitude')
        self.longitude   = weeutil.weeutil.latlon_string(self.longitude_f, 'E' if self.longitude_f>0 else 'W')
        self.altitude_f  = config_dict['Station'].as_float('altitude')
        # TODO: Shouldn't hard wire in the reference to 'ImperialUnits':
        alt_units = config_dict['Labels']['ImperialUnits']['altitude']
        self.altitude = "%d %s" % (self.altitude_f, alt_units)
        self.location        = config_dict['Station']['location']
        self.rain_year_start = int(config_dict['Station'].get('rain_year_start', '1'))
        self.radar_url       = config_dict['Station'].get('radar_url','')
        self.uptime = weeutil.weeutil.secs_to_string(time.time() - weewx.launchtime_ts) if weewx.launchtime_ts else ''
        self.version = weewx.__version__
        # The following works on Linux only:
        try:
            os_uptime_secs =  float(open("/proc/uptime").read().split()[0])
            self.os_uptime = weeutil.weeutil.secs_to_string(int(os_uptime_secs + 0.5))
        except:
            self.os_uptime = ''
    
