#
#    Copyright (c) 2009 Tom Keffer <tkeffer@gmail.com>
#
#    See the file LICENSE.txt for your full rights.
#
"""Compute statistics, manage the statistical database

    Keeps a running tally in a database of the min, max, avg of weather data
    for each day. Also, keeps some specialized data for wind.

    General strategy is that archive data is used for min, max data and
    for averages (including wind vector averages). Fast changing LOOP
    data is also used for min, max data (where it can give higher resolution)
    and for specialized, wind rms data. It is not used for averages. 
    
    The advantage of this approach is that LOOP data is easy to miss --- neither
    the weather station nor weewx stores it. Archived data is stored, so
    it's easy to perform catchups. It's generally preferred to use it, and
    for many types it's good enough. On the other hand, LOOP has much better
    resolution, and is essential for RMS calculations.
    
    Note that a date does not include midnight --- that belongs
    to the previous day. That is because a data record archives
    the *previous* interval. So, for the date 5-Oct-2008 with
    a five minute archive interval, the statistics would include
    the following records:
      5-Oct-2008 00:05:00
      5-Oct-2008 00:10:00
      5-Oct-2008 00:15:00
      .
      .
      .
      5-Oct-2008 23:55:00
      6-Oct-2008 00:00:00

    Statistics can be calculated for a day, month, year, and
    rain year.    

    This version doesn't take full advantage of polymorphism. There's
    too much special knowledge outside of the StandardStats and WindStats
    classes of their internal structure and requirements, especially
    about SQL statements.

"""

from __future__ import with_statement
import math
import time
import datetime
import os
import os.path
import syslog
from pysqlite2 import dbapi2 as sqlite3
import weeutil.weeutil


# The default types for which statistical data should be kept,
# should the user not specify anything in the config file.
# This is pretty much all possible types, which can result
# in a much bigger than necessary stats database.
default_stats_types = ('barometer',
                       'inTemp',
                       'outTemp',
                       'inHumidity',
                       'outHumidity',
                       'rainRate',
                       'rain',
                       'dewpoint',
                       'windchill',
                       'heatindex',
                       'ET',
                       'radiation',
                       'UV',
                       'extraTemp1',
                       'extraTemp2',
                       'extraTemp3',
                       'soilTemp1',
                       'soilTemp2',
                       'soilTemp3',
                       'soilTemp4',
                       'leafTemp1',
                       'leafTemp2',
                       'extraHumid1',
                       'extraHumid2',
                       'soilMoist1',
                       'soilMoist2',
                       'soilMoist3',
                       'soilMoist4',
                       'leafWet1',
                       'leafWet2',
                       'rxCheckPercent',
                       'wind')

std_create_str  = """CREATE TABLE %s   ( dateTime INTEGER NOT NULL UNIQUE PRIMARY KEY, """\
                  """min REAL, mintime INTEGER, max REAL, maxtime INTEGER, sum REAL, count INTEGER);"""

wind_create_str = """CREATE TABLE wind ( dateTime INTEGER NOT NULL UNIQUE PRIMARY KEY, """\
                  """min REAL, mintime INTEGER, max REAL, maxtime INTEGER, sum REAL, count INTEGER, """\
                  """gustdir REAL, xsum REAL, ysum REAL, squaresum REAL, squarecount INTEGER);"""

meta_create_str = """CREATE TABLE metadata (name TEXT NOT NULL UNIQUE PRIMARY KEY, value TEXT);"""
                 
std_replace_str  = """REPLACE INTO %s   VALUES(?, ?, ?, ?, ?, ?, ?)"""
wind_replace_str = """REPLACE INTO wind VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)"""
meta_replace_str = """REPLACE into metadata VALUES(?, ?)"""  

class StatsDict(dict):
    """Holds statistics for all types, from a starting time stamp

    The class inherits from a dictionary, where the key is a data type,
    and the value is an instance of StandardStats or its subclass, WindStats. 
    
    For example, temperature statistics are available as
      self['outTemp'].max
      self['outTemp'].maxtime
      self['outTemp'].avg
    """
    def __init__(self, stats_types, start_ts):
        """Initialize a new instance of StatsDict
        
        stats_types: A sequence of types for which this instance will hold statistics.
        Generally, these are the SQL types. An exception is 'wind', which is a kind of
        synthetic type, because it is more vector like. An example 
        would be: ('outTemp', 'inTemp', 'barometer', 'wind', etc.)
        
        start_ts: The starting time from which stats were accumulated for this instance.
        Typically, this is the start of a day, month, year, etc.
        """
        
        self.start_ts = start_ts
        for _stats_type in stats_types:
            # Determine the correct subclass, depending on the desired
            # type
            self[_stats_type] = StandardStats() if _stats_type != 'wind' else WindStats()
        
    def addArchiveRecord(self, rec):
        """Add an archive record to the running statistical tally for all types.
        
        rec: A station archive record
        """
        # For each type, add the archive record:
        for _stats_type in self.keys() :
            self[_stats_type].addArchiveRecord(rec, _stats_type)
    
    def addLoopRecord(self, rec):
        """Add a LOOP record to the running statistical tally for all types.
        
        rec: A station loop record
        """
        # For each type, add the archive record:
        for _stats_type in self.keys() :
            self[_stats_type].addLoopRecord(rec, _stats_type)
    
        
    def merge(self, statsDict):
        """Merge the statistics of another StatsDict into myself"""
        
        for _stats_type in self.keys():
            self[_stats_type].merge(statsDict[_stats_type])
            
class StatsDb(object):
    """Manage the sqlite3 statistical database
    
    After initialization, the attribute self.stats_types will contain a list
    of the types for which statistics are being gathered, or None if
    the database has not been initialized. This can be used to determine
    whether this is a virgin database.
    """
    def __init__(self, stats_filename):
        """Create an instance of StatsDb to manage a database.
        
        """
        self.stats_filename = stats_filename
        self.stats_types    = self.__getTypes()
        
    def addArchiveRecord(self, rec):
        time_ts = rec['dateTime']
        # Get the start-of-day for the time stamp:
        _sod_ts = weeutil.weeutil.startOfArchiveDay(time_ts)

        dayStats = self._getDayStats(_sod_ts)
        # ... and add this record to the running tally...
        dayStats.addArchiveRecord(rec)
        # ... record back in the stats database:
        self._setDayStats(dayStats, rec['dateTime'])
    
    def addLoopRecord(self, rec):
        time_ts = rec['dateTime']
        # Get the start-of-day for the time stamp:
        _sod_ts = weeutil.weeutil.startOfArchiveDay(time_ts)

        dayStats = self._getDayStats(_sod_ts)
        # ... and add this packet to the running tally...
        dayStats.addLoopRecord(rec)
        # ... record back in the stats database:
        self._setDayStats(dayStats, rec['dateTime'])
    
        
    def getAllStats(self, time_ts, rain_year_start = 1):
        """Get day, month, year, and rain year statistics and return them.
        
        time_ts: A timestamp within the day, month, year, and rain year for
        which stats are desired.
        
        rain_year_start: The month the rain year starts. [Optional. Defaults to '1']
        
        returns: a 5-way tuple of instances of StatsDict, respectively day, week, month,
        year, and rain year.
        """

        _sod_ts  = weeutil.weeutil.startOfDay(time_ts)
        _sow_ts  = weeutil.weeutil.startOfWeek(time_ts)
        _som_ts  = weeutil.weeutil.startOfMonth(time_ts)
        _soy_ts  = weeutil.weeutil.startOfYear(time_ts)
        _sory_ts = weeutil.weeutil.startOfRainYear(time_ts, rain_year_start)
        
        # Get a bunch of empty StatsDicts
        dayStats      = StatsDict(self.stats_types, _sod_ts)
        weekStats     = StatsDict(self.stats_types, _sow_ts)
        monthStats    = StatsDict(self.stats_types, _som_ts)
        yearStats     = StatsDict(self.stats_types, _soy_ts)
        rainYearStats = StatsDict(self.stats_types, _sory_ts)
        
        start_ts = min(_sod_ts, _sow_ts, _som_ts, _soy_ts, _sory_ts)
        finished = False

        # Now fill up the StatsDicts...
        for _day in weeutil.weeutil.stampgen(start_ts, time_ts, 24 * 3600) :
            assert(not finished)           # Sanity check
            stats = self._getDayStats(_day)
            if _day >= _sory_ts:
                rainYearStats.merge(stats)
            if _day >= _soy_ts:
                yearStats.merge(stats)
            if _day >= _som_ts:
                monthStats.merge(stats)
            if _day >= _sow_ts:
                weekStats.merge(stats)
            if _day == _sod_ts:
                dayStats = stats
                # If we've reached the day, we should be done. This is a little
                # sanity check:
                finished = True

        return(dayStats, weekStats, monthStats, yearStats, rainYearStats)
            
    def config(self, stats_types = None):
        """Initialize the StatsDb database
        
        Does nothing if the database has already been initialized.
        
        stats_types: an iterable collection with the names of the types for
        which statistics will be gathered [optional. Default is to use all
        possible types]
        
        """
        # Check whether the database exists:
        if os.path.exists(self.stats_filename):
            syslog.syslog(syslog.LOG_INFO, "stats: statistical database %s already exists." % self.stats_filename)
        else:
            # If it doesn't exist, create the parent directories
            archiveDirectory = os.path.dirname(self.stats_filename)
            if not os.path.exists(archiveDirectory):
                syslog.syslog(syslog.LOG_NOTICE, "stats: making archive directory %s." % archiveDirectory)
                os.makedirs(archiveDirectory)
            
        # If it has no schema, initialize it:
        if not self.stats_types:
            # No schema. Need to initialize the stats database.
            
            # If nothing has been specified, use the default stats types:
            if not stats_types:
                stats_types = default_stats_types
            
            with sqlite3.connect(self.stats_filename) as _connection:
            
                for _stats_type in stats_types:
                    if _stats_type == 'wind':
                        _connection.execute(wind_create_str)
                    else:
                        _connection.execute(std_create_str % (_stats_type,))
                _connection.execute(meta_create_str)
                
            self.stats_types = stats_types
            syslog.syslog(syslog.LOG_NOTICE, "stats: created schema for statistical database %s." % self.stats_filename)


    def backfill(self, archive, start_ts = None, stop_ts = None):
        """Fill the statistical database from the archive database.
        
        Normally, the stats database if filled by LOOP packets (to get
        maximum time resolution), but if it gets corrupted, or if a new
        user is starting up with imported wview data, it's necessary to
        recreate it from straight archive data. The Hi/Lows will all be there,
        but the times won't be any more accurate than the archive period.
        """
        t1 = time.time()
        nrecs = 0
        ndays = 0
        
        _dayStats = None
        _lastTime = None
        
        if start_ts is None:
            start_ts = self.__getLastUpdate()
        
        for _rec in archive.getBatchRecords(start_ts, stop_ts):
            _rec_time_ts = _rec['dateTime']
            _rec_sod_ts = weeutil.weeutil.startOfArchiveDay(_rec_time_ts)
            # Check whether this is the first day, or we have entered a new day:
            if _dayStats is None or _dayStats.start_ts != _rec_sod_ts:
                    # We have. Write the old day back to the database:
                    if _dayStats:
                        self._setDayStats(_dayStats, _lastTime)
                        ndays += 1
                    # Get the new day:
                    _dayStats = StatsDict(self.stats_types, _rec_sod_ts)
                
            _dayStats.addArchiveRecord(_rec)
            nrecs += 1
            _lastTime = _rec_time_ts
    
        self._setDayStats(_dayStats, _lastTime)
        ndays += 1
        
        t2 = time.time()
        tdiff = t2 - t1
        if nrecs:
            syslog.syslog(syslog.LOG_NOTICE, 
                      "stats: backfilled %d days of statistics with %d records in %.3f seconds" % (ndays, nrecs, tdiff))
        else:
            syslog.syslog(syslog.LOG_INFO, "stats: stats database up to date.")

    def _getDayStats(self, sod_ts):
        """Get the statistics for a day.
        
        sod_ts: The start-of-day in unix epoch time.
        
        returns: An instance of StatsDict
        """
        # Make sure this is truly the start of the day:
        assert(weeutil.weeutil.startOfDay(sod_ts) == sod_ts)
        
        _connection = sqlite3.connect(self.stats_filename)
        _cursor = _connection.cursor()
        
        dayStats = StatsDict(self.stats_types, sod_ts)
        
        for _stats_type in dayStats.keys():

            # Form a SQL select statement for the appropriate type
            _sql_str = "SELECT * FROM %s WHERE dateTime = ?" % (_stats_type,)
            # Peform the select, against the desired timestamp
            _cursor.execute(_sql_str, (sod_ts,))
            # Get the result
            row = _cursor.fetchone()
            # Transfer the data (if any) to the stats for the right type
            dayStats[_stats_type].setToRow(row)

        _cursor.close()
        _connection.close()
    
        return dayStats

    def _setDayStats(self, dayStats, time_ts):
        """Write the statistics for a day to the database.
        
        """
        if dayStats is None : return
        
        # Using the connection as a context manager means that
        # in case of an error, all tables will get rolled back.
        with sqlite3.connect(self.stats_filename) as _connection:
            for _stats_type in dayStats.keys():
                vals = dayStats[_stats_type].getRow()
                vals[0] = dayStats.start_ts
                
                # Need a slightly different SQL statement for the wind data:
                if _stats_type == 'wind':
                    replace_str = wind_replace_str
                else:
                    replace_str = std_replace_str % _stats_type
                
                _connection.execute(replace_str, vals)
            # Update the time of the last stats update:
            _connection.execute(meta_replace_str, ('lastUpdate', str(int(time_ts))))
                
    def __getTypes(self):
        """Query the sqlite_master database table to determine which types
        we are gathering statistics for.
        
        returns: A list of types or None if the database has not been initialized.
        
        """
        _connection = sqlite3.connect(self.stats_filename)
        _cursor = _connection.execute('''select name from sqlite_master where type = 'table';''')
        
        stats_types = [str(row[0]) for row in _cursor if row[0] != u'metadata']
        if len(stats_types) == 0 :
            stats_types = None
        
        _cursor.close()
        _connection.close()

        return stats_types
    
    def __getLastUpdate(self):
        _connection = sqlite3.connect(self.stats_filename)
        _cursor = _connection.execute("""select value from metadata where name = 'lastUpdate';""")
        
        _row = _cursor.fetchone()
        _cursor.close()
        _connection.close()
        return int(_row[0]) if _row is not None else None
    
class StandardStats(object):
    """Keeps track of the min and max encountered and when.
    It also keeps a running sum and count, so averages can
    be calculated.
    
    """ 
    def __init__(self):
        self.clearStats()
        
    def clearStats(self):
        self.min     = None
        self.max     = None
        self.mintime = None
        self.maxtime = None
        self.sum     = 0.0
        self.count   = 0
        
    def setToRow(self, row):
        """Extract data from a db row and set myself to it."""
        if row is None:
            self.clearStats()
        else:
            self.min     = row[1]
            self.mintime = row[2]
            self.max     = row[3]
            self.maxtime = row[4]
            self.sum     = row[5]
            self.count   = row[6]
    
    def getRow(self):
        """Return a db row"""
        row = [None,            # A timestamp will go here
               self.min,
               self.mintime,
               self.max,
               self.maxtime,
               self.sum,
               self.count]
        return row
    
    def addToHiLow(self, rec, stats_type):
        """Add a new record to the running hi/low tally.
        """
        v = rec.get(stats_type)
        # NB: v could be None because either there is no data for
        # this type, or this type doesn't exist at all in the record.
        if v is None :
            return
        if self.min is None or v < self.min:
            self.min = v
            self.mintime = rec['dateTime']
        if self.max is None or v > self.max:
            self.max = v
            self.maxtime = rec['dateTime']
    
    def addToSum(self, rec, stats_type):
        """Add a new record to the running sum and count.
        """
        v = rec.get(stats_type)
        
        if v is None:
            return
        self.sum   += v
        self.count += 1
        
    def addArchiveRecord(self, rec, stats_type):
        """Add an archive record to the accumulating stats.
        
        Archive data is used for the Hi/Low data, as well as
        averages.
        """
        self.addToHiLow(rec, stats_type)
        self.addToSum(rec, stats_type)
        
    def addLoopRecord(self, rec, stats_type):
        """Add a LOOP record to the accumulating stats.
        
        A LOOP record is used only for Hi/Low data for standard
        stats.
        """
        self.addToHiLow(rec, stats_type)
        
    def merge(self, stats):
        """Merge my statistics with the statistics of another instance of my type."""
        assert(isinstance(stats, StandardStats))
        if stats.min is not None:
            if self.min is None or stats.min < self.min:
                self.min     = stats.min
                self.mintime = stats.mintime
        if stats.max is not None:
            if self.max is None or stats.max > self.max:
                self.max     = stats.max
                self.maxtime = stats.maxtime
                    
        self.sum   += stats.sum
        self.count += stats.count

        
    def getAvg(self):
        return self.sum / self.count if self.count != 0 else None

    avg = property(getAvg)
    
    def __str__(self):
        """Return self as a string. Useful for diagnostics & debugging.
        """
        min_str = "%f" % self.min if self.min is not None else "N/A"
        max_str = "%f" % self.max if self.max is not None else "N/A"
        mintime_str = time.strftime("%Y-%m-%d %H:%M", time.localtime(self.mintime))
        maxtime_str = time.strftime("%Y-%m-%d %H:%M", time.localtime(self.maxtime))
        avg_str = "%f" % self.avg if self.avg is not None else "N/A"
        sum_str = "%f" % self.sum if self.sum is not None else "N/A"
        return "min = %s (%s); max = %s (%s); avg = %s; sum = %s; " % (min_str, mintime_str, max_str, maxtime_str, avg_str, sum_str) 
    
class WindStats(StandardStats):
    """Specialized version of StandardStats for wind data.
    
    """
    def __init__(self):
        WindStats.clearStats(self)
        
    def clearStats(self):
        super(WindStats, self).clearStats()
        self.gustdir     = None
        self.xsum        = 0.0
        self.ysum        = 0.0
        self.squaresum   = 0.0
        self.squarecount = 0
        
    def setToRow(self, row):
        if row is None:
            self.clearStats()
        else :
            super(WindStats, self).setToRow(row)
            self.gustdir     = row[7]
            self.xsum        = row[8]
            self.ysum        = row[9]
            self.squaresum   = row[10]
            self.squarecount = row[11]
    
    def getRow(self):
        """Get a db row from myself."""
        row = super(WindStats, self).getRow()
        row.append(self.gustdir)
        row.append(self.xsum)
        row.append(self.ysum)
        row.append(self.squaresum)
        row.append(self.squarecount)
        return row
        
    def addToHiLow(self, rec, stats_type):
        """Specialized version for wind data. It differs from
        the standard addToHiLow in that it takes advantage of 
        wind gust data if it's available. It also keeps track
        of the *direction* of the high wind, as well as its time 
        and magnitude.
        """
        assert(stats_type == 'wind')
        # Get the wind speed & direction, breaking them down into vector
        # components.
        v      = rec.get('windSpeed')
        vHi    = rec.get('windGust')
        vHiDir = rec.get('windGustDir')
        if vHi is None:
            # This typically happens because a LOOP packet is being added, which
            # does not have windGust data.
            vHi    = v
            vHiDir = rec.get('windDir')

        if v is not None and (self.min is None or v < self.min):
            self.min = v
            self.mintime = rec['dateTime']
        if vHi is not None and (self.max is None or vHi > self.max):
            self.max = vHi
            self.maxtime = rec['dateTime']
            self.gustdir = vHiDir
    
    def addToSum(self, rec, stats_type):
        """Specialized version for wind data. It differs from
        the standard addToSum in that it calculates a vector
        average as well. """
        assert(stats_type == 'wind')
        # Get the wind speed & direction, breaking them down into vector
        # components.
        speed = rec.get('windSpeed')
        theta = rec.get('windDir')
        if speed is not None:
            self.sum   += speed
            self.count += 1
            if theta is not None :
                self.xsum      += speed * math.cos(math.radians(90.0 - theta))
                self.ysum      += speed * math.sin(math.radians(90.0 - theta))
    
    def addToRms(self, rec, stats_type):
        """Add a record to the wind-specific rms stats"""
        assert(stats_type == 'wind')
        speed = rec.get('windSpeed')
        if speed is not None:
            self.squaresum   += speed**2
            self.squarecount += 1

    def addLoopRecord(self, rec, stats_type):
        """Specialized version for wind data.
        
        Same as the version for StandardStats, except it also updates the RMS data
        """
        self.addToHiLow(rec, stats_type)
        self.addToRms(rec, stats_type)
        
    def merge(self, stats):
        """Merge my statistics with another instance of my type
        """
        assert(isinstance(stats, WindStats))    # Sanity check
        if stats.min is not None:
            if self.min is None or stats.min < self.min:
                self.min     = stats.min
                self.mintime = stats.mintime
        if stats.max is not None:
            if self.max is None or stats.max > self.max:
                self.max      = stats.max
                self.maxtime  = stats.maxtime
                self.gustdir  = stats.gustdir
                    
        self.sum         += stats.sum
        self.count       += stats.count
        self.xsum        += stats.xsum
        self.ysum        += stats.ysum
        self.squaresum   += stats.squaresum
        self.squarecount += stats.squarecount
        
    def getRms(self):
        return math.sqrt(self.squaresum / self.squarecount) if self.squarecount != 0 else None
    def getVecAvg(self):
        return math.sqrt((self.xsum**2 + self.ysum**2) / self.count**2) if self.count != 0 else None
    def getVecDir(self):
        deg = 90.0 - math.degrees(math.atan2(self.ysum, self.xsum))
        return deg if deg > 0 else deg + 360.0
    
    rms    = property(getRms)
    vecavg = property(getVecAvg)
    vecdir = property(getVecDir)
    
    def __str__(self):
        """Useful for diagnostics and debugging.
        
        """
        rms_str = "%f" % self.rms if self.rms is not None else "N/A"
        vecavg_str = "%f" % self.vecavg if self.vecavg is not None else "N/A"
        vecdir_str = "%f" % self.vecdir if self.vecdir is not None else "N/A"
        return super(WindStats, self).__str__() + "; rms = %s; vecavg = %s; vecdir = %s " % (rms_str, vecavg_str, vecdir_str)

    
if __name__ == '__main__':
    import configobj
    
    def test(config_path):
        
        try :
            config_dict = configobj.ConfigObj(config_path, file_error=True)
        except IOError:
            print "Unable to open configuration file ", config_path
            exit()
            
        filename = os.path.join(config_dict['Station']['WEEWX_ROOT'], 
                                config_dict['Stats']['stats_file'])
        stats = StatsDb(filename)
        stats.config(config_dict['Stats'].get('stats_types'))
        
        time_ts = 1240624930
        t1 = time.time()
        (dayStats, weekStats, monthStats, yearStats, rainYearStats) = stats.getAllStats(time_ts, 10)
        t2 = time.time()
        print "Elapsed time = %f" % (t2-t1)
        print "********* barometer data ********"
        print dayStats['barometer']
        print weekStats['barometer']
        print monthStats['barometer']
        print yearStats['barometer']
        print rainYearStats['barometer']
        
        print "\n********* wind data ********"
        print dayStats['wind']
        print weekStats['wind']
        print monthStats['wind']
        print yearStats['wind']
        print rainYearStats['wind']
        
    test('/home/weewx/weewx.conf')
        